@@ -6,6 +6,7 @@ import collections.abc
 import datetime
 import functools
 import itertools
+import re
 import warnings
 
 from collections import deque
@@ -48,50 +49,86 @@ Namespace = Mapping[str, object]
 _undefined = object()
 
 
+# pylint: disable=too-many-return-statements
 def _getitem(obj: Any, key: Any) -> Any:
     """Item getter with special methods for arrays/lists and hashes/dicts."""
     # NOTE: A runtime checkable protocol was too slow.
     if hasattr(key, "__liquid__"):
         key = key.__liquid__()
 
-    if key == "size" and isinstance(obj, collections.abc.Sized):
-        return len(obj)
-    if (
-        key == "first"
-        and not isinstance(obj, str)
-        and isinstance(obj, collections.abc.Sequence)
-    ):
-        return obj[0]
-    if (
-        key == "last"
-        and not isinstance(obj, str)
-        and isinstance(obj, collections.abc.Sequence)
-    ):
-        return obj[-1]
+    if key == "size":
+        try:
+            return getitem(obj, "size")
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Sized):
+                return len(obj)
+            raise
+
+    if key == "first" and not isinstance(obj, str):
+        try:
+            return getitem(obj, "first")
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Mapping) and obj:
+                return list(itertools.islice(obj.items(), 1))[0]
+            if isinstance(obj, collections.abc.Sequence):
+                return obj[0]
+            raise
+
+    if key == "last" and not isinstance(obj, str):
+        try:
+            return getitem(obj, "last")
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Sequence):
+                return obj[-1]
+            raise
 
     return getitem(obj, key)
 
 
+# pylint: disable=too-many-return-statements
 async def _getitem_async(obj: Any, key: Any) -> object:
     """Item getter with special methods for arrays/lists and hashes/dicts."""
     # NOTE: A runtime checkable protocol was too slow.
     if hasattr(key, "__liquid__"):
         key = key.__liquid__()
 
-    if key == "size" and isinstance(obj, collections.abc.Sized):
-        return len(obj)
-    if (
-        key == "first"
-        and not isinstance(obj, str)
-        and isinstance(obj, collections.abc.Sequence)
-    ):
-        return obj[0]
-    if (
-        key == "last"
-        and not isinstance(obj, str)
-        and isinstance(obj, collections.abc.Sequence)
-    ):
-        return obj[-1]
+    if key == "size":
+        try:
+            return (
+                await obj.__getitem_async__(key)
+                if hasattr(obj, "__getitem_async__")
+                else getitem(obj, "size")
+            )
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Sized):
+                return len(obj)
+            raise
+
+    if key == "first" and not isinstance(obj, str):
+        try:
+            return (
+                await obj.__getitem_async__(key)
+                if hasattr(obj, "__getitem_async__")
+                else getitem(obj, "first")
+            )
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Mapping) and obj:
+                return list(itertools.islice(obj.items(), 1))[0]
+            if isinstance(obj, collections.abc.Sequence):
+                return obj[0]
+            raise
+
+    if key == "last" and not isinstance(obj, str):
+        try:
+            return (
+                await obj.__getitem_async__(key)
+                if hasattr(obj, "__getitem_async__")
+                else getitem(obj, "last")
+            )
+        except (KeyError, IndexError, TypeError):
+            if isinstance(obj, collections.abc.Sequence):
+                return obj[-1]
+            raise
 
     if hasattr(obj, "__getitem_async__"):
         return await obj.__getitem_async__(key)
@@ -318,6 +355,7 @@ class Context:
         "disabled_tags",
         "autoescape",
         "_copy_depth",
+        "parent_context",
     )
 
     getitem = _getitem
@@ -329,6 +367,7 @@ class Context:
         globals: Optional[Namespace] = None,
         disabled_tags: Optional[List[str]] = None,
         copy_depth: int = 0,
+        parent_context: Optional[Context] = None,
     ):
         self.env = env
 
@@ -347,7 +386,7 @@ class Context:
         # temporary namespace is pushed to the front of this chain.
         self.scope = ReadOnlyChainMap(self.locals, self.globals, builtin, self.counters)
 
-        # A namspace supporting stateful tags. Such as `cycle`, `increment`,
+        # A namespace supporting stateful tags. Such as `cycle`, `increment`,
         # `decrement` and `ifchanged`.
         self.tag_namespace: Dict[str, Any] = {
             "cycles": {},
@@ -371,6 +410,9 @@ class Context:
         # gracefully.
         self._copy_depth = copy_depth
 
+        # A reference to a parent render context, if this one has been copied.
+        self.parent_context = parent_context
+
     def assign(self, key: str, val: Any) -> None:
         """Add `val` to the context with key `key`."""
         self.locals[key] = val
@@ -378,11 +420,11 @@ class Context:
     def get(self, path: ContextPath, default: object = _undefined) -> object:
         """Return the value at path `path` if it is in scope, else default."""
         if isinstance(path, str):
-            return self.resolve(path, default)
+            return self._resolve(path, default)
 
         name, items = path[0], path[1:]
         assert isinstance(name, str)
-        obj = self.resolve(name, default)
+        obj = self._resolve(name, default)
 
         if items:
             try:
@@ -409,11 +451,11 @@ class Context:
     ) -> object:
         """Return the value at path `path` if it is in scope, else default."""
         if isinstance(path, str):
-            return self.resolve(path, default)
+            return self._resolve(path, default)
 
         name, items = path[0], path[1:]
         assert isinstance(name, str)
-        obj = self.resolve(name, default)
+        obj = self._resolve(name, default)
 
         if items:
             _gi = Context.getitem_async
@@ -435,6 +477,9 @@ class Context:
         This is like `get`, but does a single, top-level lookup rather than a
         chained lookup from a sequence of keys.
         """
+        return self._resolve(name, default)
+
+    def _resolve(self, name: str, default: object = _undefined) -> Any:
         try:
             return self.scope[name]
         except KeyError:
@@ -559,17 +604,18 @@ class Context:
         self, namespace: Namespace, disabled_tags: Optional[List[str]] = None
     ) -> Context:
         """Return a copy of this context without any local variables or other state
-        for statefull tags."""
+        for stateful tags."""
         if self._copy_depth > MAX_CONTEXT_DEPTH:
             raise ContextDepthError(
                 "maximum context depth reached, possible recursive render"
             )
 
-        return Context(
+        return type(self)(
             self.env,
             globals=ReadOnlyChainMap(namespace, self.globals),
             disabled_tags=disabled_tags,
             copy_depth=self._copy_depth + 1,
+            parent_context=self,
         )
 
     def error(self, exc: Error) -> None:
@@ -578,3 +624,81 @@ class Context:
             raise exc
         if self.env.mode == Mode.WARN:
             warnings.warn(str(exc), category=lookup_warning(exc.__class__))
+
+
+class VariableCaptureContext(Context):
+    """A render context that captures template variable names."""
+
+    # Used for formatting context path strings.
+    re_ident = re.compile(r"^[\w_][\w_\-]*$")
+
+    def __init__(
+        self,
+        env: Environment,
+        globals: Optional[Namespace] = None,
+        disabled_tags: Optional[List[str]] = None,
+        copy_depth: int = 0,
+        parent_context: Optional[VariableCaptureContext] = None,
+    ):
+        super().__init__(env, globals, disabled_tags, copy_depth, parent_context)
+        self.local_references: List[str] = []
+        self.all_references: List[str] = []
+        self.undefined_references: List[str] = []
+
+        root_context: VariableCaptureContext = self
+        while root_context.parent_context and isinstance(
+            root_context.parent_context, VariableCaptureContext
+        ):
+            root_context = root_context.parent_context
+        self.root_context = root_context
+
+    def assign(self, key: str, val: Any) -> None:
+        self.local_references.append(key)
+        return super().assign(key, val)
+
+    def get(self, path: ContextPath, default: object = _undefined) -> object:
+        result = super().get(path, default)
+        self._count_reference(path, result)
+        return result
+
+    async def get_async(
+        self, path: ContextPath, default: object = _undefined
+    ) -> object:
+        result = await super().get_async(path, default)
+        self._count_reference(path, result)
+        return result
+
+    def resolve(self, name: str, default: object = _undefined) -> Any:
+        result = super().resolve(name, default)
+        self._count_reference(name, result)
+        return result
+
+    def increment(self, name: str) -> int:
+        self.local_references.append(name)
+        return super().increment(name)
+
+    def decrement(self, name: str) -> int:
+        self.local_references.append(name)
+        return super().decrement(name)
+
+    def _count_reference(self, path: ContextPath, result: object) -> None:
+        if isinstance(path, str):
+            ref = path
+        else:
+            _path = []
+            for elem in path:
+                if isinstance(elem, int):
+                    _path.append(f"[{elem}]")
+                else:
+                    if self.re_ident.match(elem):
+                        if _path:
+                            _path.append(f".{elem}")
+                        else:
+                            _path.append(f"{elem}")
+                    else:
+                        _path.append(f'["{elem}"]')
+            ref = "".join(_path)
+
+        if is_undefined(result):
+            self.root_context.undefined_references.append(ref)
+        self.root_context.all_references.append(ref)
