@@ -82,7 +82,7 @@ class FilterError(Error):
 
 
 class FilterArgumentError(Error):
-    """Exception raised when a filters arguments are invalid."""
+    """Exception raised when a filter's arguments are invalid."""
 
 
 class FilterValueError(Error):
@@ -90,7 +90,7 @@ class FilterValueError(Error):
 
 
 class TemplateNotFound(Error):
-    """Excpetions raised when a template could not be found."""
+    """Excpetion raised when a template could not be found."""
 
     def __str__(self) -> str:
         msg = super().__str__()
@@ -117,6 +117,10 @@ class ContinueLoop(LiquidInterrupt):
     """Exception raised when a ContinueNode is rendered."""
 
 
+class TemplateTraversalError(Error):
+    """Exception raised when an AST node or expression can not be visited."""
+
+
 class LiquidWarning(UserWarning):
     """Base warning."""
 
@@ -163,7 +167,7 @@ class Markup(str):
             "autoescape requires Markupsafe to be installed"
         )  # pragma: no cover
 
-    # pylint: disable=no-self-use,missing-function-docstring
+    # pylint: disable=missing-function-docstring
     def unescape(self) -> str:
         raise Error(
             "autoescape requires Markupsafe to be installed"
